from __future__ import annotations  # PEP 604 backport

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from ._logger import get_logger

sns.set_theme(style="ticks")

PADDING = 0.10


class Visualizer:
    """
    `Visualizer` includes methods for plotting the estimated loss gradient, the efficient frontier,
    and L-BFGS optimization results. It helps in understanding the distribution and relationships between
    compute resources, model parameters, and data samples, and highlights efficient allocation frontiers
    and seed regimes.

    Attributes:
        project_dir (str): Directory to save plots.
        logger (Logger): Logger instance for visualization messages.
        cc (Chinchilla): A Chinchilla instance passed through the `plot` method to be shared across its submethods.
    """

    def __init__(self, project_dir: str, log_level: int = 30) -> None:
        """
        Constructs a Visualizer instance with a project directory and logging level.

        Args:
            project_dir (str): The directory where the plots will be saved.
            log_level (int): Logging level to control the verbosity of log messages.
        """
        self.project_dir = project_dir
        self.logger = get_logger(log_level, name="chinchilla.visualizer")

    def plot(
        self,
        cc,
        next_point: dict[str, float] | None = None,
        fitted: bool = True,
        img_name: str = "parametric_fit",
        cmap_name: str = "plasma",
        simulation: bool = False,
    ) -> None:
        """
        Plots the loss gradient and efficient frontier for resource allocation.

        This method visualizes the distribution and relationships between compute resources
        (FLOPs), model parameters, and data samples. It shows how these factors interact
        with the loss function and highlights efficient allocation frontiers and seed regimes.

        **Example output**:
        ![](../examples/efficientcube-1e15_1e16/parametric_fit.png)

        Args:
            cc: A Chinchilla instance with a Database of training runs and scaling law parameters if estimated.
            next_point (dict[str, float] | None): The next point to be plotted, if any.
            fitted (bool): Whether to plot the scaling law gradient or only raw data points.
                If the loss predictor is not fitted, falls back to False.
            img_name (str): The name of the image file to save the plot as.
            cmap_name (str): The name of the colormap to be used for plotting.
            simulation (bool): Whether the plot is for a simulation.
        """
        # Validate inputs
        if not isinstance(cc.database.df, pd.DataFrame):
            raise TypeError("The 'df' argument must be a pandas DataFrame.")
        if next_point and isinstance(next_point, dict) and len(next_point) != 3:
            raise TypeError("The 'next_point' must be None or a `dict` of (C, N, D).")
        # Ignore call
        if len(cc.database.df) < 3:
            self.logger.warning("Not enough samples in the database. Terminating plot")
            return

        # Overrides
        if fitted:
            if not all(hasattr(cc, param) for param in ["E", "A", "B", "alpha", "beta"]):
                self.logger.warning(
                    "The scaling law parameters has not been estimated. Plotting only raw data points."
                )
                fitted = False
        if not fitted:
            img_name = "raw_data_points"
        if simulation:
            img_name = f"simulation--{img_name}"
        img_filepath = os.path.join(self.project_dir, img_name + ".png")

        # set a shortcut
        self.cc = cc
        self._next_point = next_point
        self.cmap = plt.colormaps[cmap_name]
        df = self.cc.database.df

        # Pad the all points & marginalize the higher edge of the gradient
        margin = 0.20
        loss_min, loss_max = df.loss.min(), df.loss.max()
        loss_range = loss_max - loss_min
        if self._next_point:
            loss_min = min(loss_min, cc.L(self._next_point["N"], self._next_point["D"]))
        iso_losses = np.linspace(loss_min - loss_range * margin, loss_max + loss_range * margin, 32)

        fig, axes = plt.subplots(1, 3, figsize=(15, 5))
        fig.tight_layout(pad=4.0, w_pad=3.0)

        # Execute helper methods by each subplot
        for i, (x, y) in enumerate([("C", "N"), ("C", "D"), ("N", "D")]):
            ax = axes[i]
            # Get the highest value to include for each axis
            x_max, y_max = [
                max(
                    self.cc.seed_ranges[k][1],
                    self.cc.database.df[k].max(),
                    self._next_point[k] if self._next_point else -float("inf"),
                )
                for k in [x, y]
            ]

            # Adjust subplot and get the bound
            xlim, ylim = self._adjust_subplot(ax, x, y, x_max, y_max)

            # Emphasize seeds
            self._shadow_seed_regime(ax, x, y)

            # Scatter plot of `df`
            self._scatter_data_points(ax, x, y, iso_losses)

            # Visualize the scaling law derived from `self.cc.fit`.
            if fitted:
                # Loss gradient
                self._plot_loss_gradient(ax, x, y, iso_losses, y_max)

                # Efficient frontier line on the gradient
                self._draw_efficient_frontier(ax, x, y, xlim, ylim)

                # show `next_point` along with a guideline if not None,
                self._show_next(ax, x, y)

            ax.legend(loc="upper left").set_zorder(1000)

        # Add a colorbar to the right
        self._add_colorbar_to_plot(fig, axes, iso_losses)

        # Finish
        plt.suptitle("Efficient frontier on iso-loss gradient" if fitted else "Distribution of seed models", y=0.925)
        plt.savefig(img_filepath)
        plt.show()
        plt.close()

        self.logger.info(f"Image saved to [u]{img_filepath}[/]")

    def LBFGS(
        self,
        y_pred: np.ndarray,
        y_true: np.ndarray,
        C: np.ndarray | None = None,
        simulation: bool = False,
        img_name: str = "LBFGS",
    ) -> None:
        """
        Plots the results of L-BFGS optimization, including the loss history and prediction accuracy.
        This method visualizes the predicted values versus the true labels and the error distribution.

        **Example output**:
        ![](../examples/efficientcube-1e15_1e16/LBFGS.png)

        Args:
            y_pred (np.ndarray): Predicted values by the model.
            y_true (np.ndarray): True labels or values for comparison with predictions.
            C (np.ndarray | None, optional): Compute resources (FLOPs) associated with each prediction, if available.
            simulation (bool, optional): Whether the plot is for a simulation.
            img_name (str, optional): The name of the image file to save the plot as.
        """
        if simulation:
            img_name = f"simulation--{img_name}"

        margin = 0.10

        vrange = min(y_pred.min(), y_true.min()), max(y_pred.max(), y_true.max())
        self.logger.debug(f"Loss range: {vrange}")
        lims = vrange[0] - margin * (vrange[1] - vrange[0]), vrange[1] + margin * (vrange[1] - vrange[0])
        self.logger.debug(f"Loss range [u]for plotting[/]: {lims}")

        fig, axs = plt.subplots(1, 2, figsize=(8, 4))
        fig.tight_layout(pad=4.0, w_pad=3.0)
        axs[0].scatter(y_pred, y_true)
        axs[0].plot(lims, lims, ls=":", lw=1, c="black")
        axs[0].set_xlabel("Predictions")
        axs[0].set_ylabel("True labels")
        axs[0].set_xlim(*lims)
        axs[0].set_ylim(*lims)
        axs[0].set_title("Best predictions vs. True labels")

        err = y_true - y_pred
        if C is None:
            densities, bins, _ = axs[1].hist(err, density=True)
            x = np.linspace(0, err.max(), 100)
            # Check if we can fit the errors to a exponential distribution
            lmd = 1 / err.mean()
            if lmd > 0:
                # A quick test to see the goodness of fit
                from scipy.stats import ks_2samp

                cdf_data = np.cumsum(densities) / np.sum(densities)
                cdf_model = 1 - np.exp(-lmd * x)
                KS, p = ks_2samp(cdf_data, cdf_model)
                if p > 0.05:  # My apologies for using not just p-value but also the threshold.
                    self.logger.info(rf"Goodness-of-fit to Exp(λ={lmd:.2f}): {KS=}, {p=}")
                    axs[1].plot(x, lmd * np.exp(-lmd * x), ls=":", lw=1, c="black", label=rf"$λ={lmd:.2f}$")
                    axs[1].legend(loc="upper right")
            axs[1].set_ylim(0, densities.max() * 1.05)
            axs[1].set_xlabel(r"Error ($L-\hat{L}$)")
            axs[1].set_ylabel("Probability density")
            axs[1].set_title("Error distribution")
        else:
            axs[1].scatter(C, np.abs(err))
            axs[1].set_xlabel("Compute (FLOPs)")
            axs[1].set_ylabel("Absolute error")
            axs[1].set_xscale("log")
            axs[1].set_yscale("log")
            axs[1].set_title("Compute and absolute error")

        plt.suptitle("L-BFGS results")
        plt.savefig(os.path.join(self.project_dir, img_name + ".png"))
        plt.show()
        plt.close()

    def _plot_loss_gradient(self, ax, x, y, iso_losses, y_max):
        """Helper method to plot the loss gradient."""
        # / 1 for converting to float when int
        log_ymin = np.log10(self.cc.seed_ranges[y][0] / 1)
        log_ymax = np.log10(y_max / 1)
        log_ymin -= (log_ymax - log_ymin) * PADDING
        log_ymax += (log_ymax - log_ymin) * PADDING

        y_values = np.logspace(log_ymin, log_ymax, 1000, dtype=np.double)
        assert not np.isnan(y_values).sum(), (f"{100*np.isnan(y_values).mean()}% NaN:", y_values)
        for j, L in enumerate(iso_losses):
            if y == "N":
                N = y_values
                _ = self.cc.B / (L - (self.cc.E + np.exp(np.log(self.cc.A) - self.cc.alpha * np.log(N))))
                # `_` can have negatives where `L < E + A/(N^alpha)` or `L < E + B/(D^beta)`, which is invalid by design.
                # Here, we arbitrarily leave potentially negative values so as to replace as NaN with `np.power`
                with np.errstate(invalid="ignore"):
                    D = np.power(_, 1 / self.cc.beta)
            elif y == "D":
                D = y_values
                _ = self.cc.A / (L - (self.cc.E + np.exp(np.log(self.cc.B) - self.cc.beta * np.log(D))))
                # Save as for N
                with np.errstate(invalid="ignore"):
                    N = np.power(_, 1 / self.cc.alpha)
            else:
                raise NotImplementedError("Please implement it yourself.")

            C = 6 * N * D
            x_values = locals()[x]
            ax.plot(x_values, y_values, c=self.cmap(j / len(iso_losses)), zorder=1)

    def _shadow_seed_regime(self, ax, x, y, resolution: int = 100):
        """Helper method to fill the seed regime with gray."""
        if x == "C":
            c = np.logspace(*np.log10(self.cc.seed_ranges[x]), resolution)
            if y == "N":
                y_lower = np.sqrt(c / (6 * self.cc.seed_ranges.N_to_D[1]))
                y_upper = np.sqrt(c / (6 * self.cc.seed_ranges.N_to_D[0]))
            else:  # y == "D"
                y_lower = np.sqrt(c * self.cc.seed_ranges.N_to_D[0] / 6)
                y_upper = np.sqrt(c * self.cc.seed_ranges.N_to_D[1] / 6)
            x_values = c
        else:  # x in ["N", "D"]
            n = np.logspace(*np.log10(self.cc.seed_ranges[x]), resolution)
            y_lower = np.maximum(self.cc.seed_ranges["C"][0] / (6 * n), self.cc.seed_ranges.N_to_D[0] * n)
            y_upper = np.minimum(self.cc.seed_ranges["C"][1] / (6 * n), self.cc.seed_ranges.N_to_D[1] * n)
            x_values = n

        ax.fill_between(x_values, y_lower, y_upper, color="silver", alpha=0.5, zorder=0, label="Seed")

    def _adjust_subplot(self, ax, x, y, x_max, y_max):
        """Adjusts the subplot configurations and return the bound of values."""
        # Get a tuple of value range for each axis, letting the minimum of a seed range the lowest and padding the all data points
        limits_by_k = {}
        for k in [x, y]:
            # Converting to `float` in case of int dtype (`np.log` cannot intake astronomically large integers)
            max_value = {x: x_max, y: y_max}[k] / 1
            log_range = np.log(max_value) - np.log(self.cc.seed_ranges[k][0])
            limits_by_k[k] = (
                self.cc.seed_ranges[k][0] * np.exp(-log_range * PADDING),
                max_value * np.exp(log_range * PADDING),
            )

        # Apply the bounds
        xlim, ylim = limits_by_k[x], limits_by_k[y]
        if np.isnan(xlim).any():
            raise ValueError(f"NaN value(s) found for the x axis {x} bound: {xlim}")
        if np.isnan(ylim).any():
            raise ValueError(f"NaN value(s) found for the x axis {y} bound: {ylim}")
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)

        ax.set_xscale("log")
        ax.set_yscale("log")
        to_label = {"C": "FLOPs", "D": "Samples", "N": "Parameters"}
        ax.set_xlabel(f"{to_label[x]} (${x}$)")
        ax.set_ylabel(f"{to_label[y]} (${y}$)")

        return xlim, ylim

    def _scatter_data_points(self, ax, x, y, iso_losses):
        ax.scatter(
            self.cc.database.df[x],
            self.cc.database.df[y],
            c=self.cc.database.df.loss,
            vmin=iso_losses.min(),
            vmax=iso_losses.max(),
            lw=1,
            edgecolor="black",
            cmap=self.cmap,
            zorder=100,
        )

    def _show_next(self, ax, x, y):
        if self._next_point:
            ax.plot(
                self._next_point[x],
                self._next_point[y],
                "*",
                markersize=18,
                markeredgecolor="black",
                label="Next",
                zorder=1000,
            )
            kwargs = dict(ls=":", lw=1, c="black")
            # ax.axvline(self._next_point[x], ymax=self._next_point[y], **kwargs)
            # ax.axhline(self._next_point[y], xmax=self._next_point[x], **kwargs)
            ax.plot([self._next_point[x]] * 2, [0, self._next_point[y]], **kwargs)
            ax.plot([0, self._next_point[x]], [self._next_point[y]] * 2, **kwargs)

    def _draw_efficient_frontier(self, ax, x, y, xlim, ylim):
        if x == "C":
            c_lim = xlim
        elif y == "C":
            c_lim = ylim
        else:
            c_lim = (6 * xlim[0] * ylim[0], 6 * xlim[1] * ylim[1])

        allocations = np.array([self.cc.allocate_compute(C) for C in c_lim])
        efficient_frontier = {"C": c_lim, "N": allocations[:, 0], "D": allocations[:, 1]}
        ax.plot(efficient_frontier[x], efficient_frontier[y], ls=":", c="blue", zorder=10, label="Efficient frontier")

    def _add_colorbar_to_plot(self, fig, axes, iso_losses):
        ax = axes[-1]
        fig.subplots_adjust(right=0.875)
        cax = fig.add_axes(
            (0.9, ax.get_position().y0, 0.015, ax.get_position().y1 - ax.get_position().y0)
            # (left, bottom, width, height)
        )
        cbar = plt.cm.ScalarMappable(cmap=self.cmap)
        cbar.set_array([iso_losses])
        fig.colorbar(cbar, cax=cax)
        cax.set_ylabel("Loss", rotation=270, labelpad=16)



from __future__ import annotations  # PEP 604 backport

import itertools
import multiprocessing
import os
import random
import textwrap
from collections.abc import Callable

import numpy as np
from attrdictx import AttrDict
from ordinal import ordinal
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn, TimeElapsedColumn, TimeRemainingColumn
from scipy import optimize as sciop  # import fmin_l_bfgs_b  # , minimize

from ._logger import get_logger
from ._metrics import asymmetric_mae
from ._utils import is_between, search_model_config
from ._validator import ModelSearchConfig, ParamGrid, SeedRanges
from .database import Database  # Used for initializing database connections
from .visualizer import Visualizer  # Required for setting up visualization properties

# 128bit/96bit: more precise than 64bit  at the cost of approx. 2x more time
DTYPE = np.longdouble
# 64bit: yields a *slightly different*, plausibly less precise result;
# Recommendable exclusively for agile testing
# DTYPE = np.double

# Clip values by lower precision for stability with `loss_fn` and `weight_fn`.
FLOAT_TINY = np.finfo(np.single).tiny
FLOAT_LOGMAX = np.log(np.finfo(np.single).max)


class Chinchilla:
    """
    Estimates the scaling law for a deep learning task.
    Provides functionalities to:

    1. Sample models from a specified "seed" regime.
    2. Fit the loss predictor $L(N, D)$.
    3. Suggest an allocation of scaled compute.

    This module includes the `Chinchilla` class, which provides methods for sampling model configurations,
    fitting the parametric loss predictor, suggesting allocations for scaled compute budgets, etc.
    It operates in a numerical precision of **128-bit** by default and integrates with
    [`chinchilla.Database`](#chinchilladatabaseDatabase) and
    [`chinchilla.Visualizer`](#chinchillavisualizerVisualizer)
    for storing and plotting data.
    """

    E: float
    A: float
    B: float
    alpha: float
    beta: float

    def __init__(
        self,
        project_dir: str,
        param_grid: dict[str, np.ndarray | list | tuple],
        seed_ranges: dict[str, np.ndarray | list | tuple],
        model_search_config: dict[str, Callable | dict] | None = None,
        loss_fn: Callable = asymmetric_mae,  # Fits to the floor (\approx. lower bound) of the distribution $L(N, D)$
        weight_fn: Callable | None = None,  # You nay weight loss prediction errors with any input
        num_seeding_steps: int | None = None,
        scaling_factor: float | None = None,
        log_level: int | str = 20,  # logging.INFO
    ) -> None:
        """
        Initializes a Chinchilla instance with the given parameters and sets up the scaling process.

        Args:
            project_dir (str): The directory path for the project where the database and visualizations will be stored.
            param_grid (dict[str, np.ndarray | list | tuple]): A dictionary specifying the grid of parameters to search over.
            seed_ranges (dict[str, np.ndarray | list | tuple]): A dictionary specifying the ranges for seeding the model configurations.
            model_search_config (dict[str, Callable | dict] | None, optional): Configuration for model search. Defaults to None.
            loss_fn (Callable, optional): The loss function to be used for fitting. Defaults to asymmetric_mae.
            weight_fn (Callable | None, optional): A function to weight loss prediction errors. Defaults to None.
            num_seeding_steps (int | None, optional): The number of seeding steps to perform. Defaults to None.
            scaling_factor (float | None, optional): The scaling factor to be used when scaling up compute. Defaults to None.
            log_level (int | str, optional): Specifies the threshold for logging messages.
                A value of 30 suppresses standard messages while any larger values hide all messages entirely. Defaults to 20 (`logging.INFO`).

        Raises:
            ValueError: If `project_dir` does not exist or is not a directory.
            TypeError: If `loss_fn` or `weight_fn` is not callable.
            FileExistsError: If a file with the same name as `project_dir` already exists.
        """
        if not project_dir:
            raise ValueError("Please specify a directory. If it does not exist, it will be created.")

        self.logger = get_logger(log_level, name="chinchilla")

        # input validation
        ParamGrid(**param_grid)
        SeedRanges(**seed_ranges)
        if model_search_config:
            ModelSearchConfig(**model_search_config)
        else:
            self.logger.warning(
                textwrap.dedent(
                    """\
                You did not specify `model_search_config`. In order to find a model configuration, You will need to either:
                1.  [b]Assign `model_search_config` attribute[/] [i]before[/] calling `seed` method.
                2.  Find the model configuration closest to `N` yourself, and optionally call `cc.adjust_D_to_N(N)` when scaling."""
                )
            )
        if not callable(loss_fn):
            raise TypeError("`loss_fn` must be callable")
        if weight_fn and not callable(loss_fn):
            raise TypeError("`weight_fn` must be callable or None")

        # Convert dict to AttrDict for easy access
        seed_ranges = AttrDict(seed_ranges)

        """Initialize configurations"""
        # Seed
        self.seed_ranges = AttrDict(
            # User-specified
            C=[float(c) for c in seed_ranges.C],  # tuple/list of large integers (>2 ** 63) can result in errors
            N_to_D=seed_ranges.N_to_D,
            # Pre-compute the bounds of allocations for the seed models
            N=[
                np.sqrt(seed_ranges.C[0] / (6 * seed_ranges.N_to_D[1])),  # lower bound
                np.sqrt(seed_ranges.C[1] / (6 * seed_ranges.N_to_D[0])),  # upper bound
            ],
            D=[
                np.sqrt(seed_ranges.C[0] * seed_ranges.N_to_D[0] / 6),  # lower bound
                np.sqrt(seed_ranges.C[1] * seed_ranges.N_to_D[1] / 6),  # upper bound
            ],
        )

        # Fit
        self.model_search_config = model_search_config
        self.param_grid = param_grid
        self.loss_fn = loss_fn
        self.weight_fn = weight_fn

        # Scale (optionally preset, optionally overridden)
        self.num_seeding_steps = num_seeding_steps  # Only for when using `step` shorthand
        self.scaling_factor = scaling_factor

        """Initialize/load the scaling database & visualizer"""
        self.project_dir = project_dir
        if os.path.isfile(self.project_dir):
            raise FileExistsError(f'There is a file matching the specified directory name "{self.project_dir}"')
        elif not os.path.exists(self.project_dir):
            self.logger.info(f"Creating a directory: [u]{project_dir}[/]")
            os.mkdir(self.project_dir)
        else:
            self.logger.info(f"Mounting a directory: [u]{self.project_dir}[/]")

        self.database = Database(self.project_dir, log_level=log_level)
        self.visualizer = Visualizer(self.project_dir, log_level=log_level)
        self._create_shortcuts()
        self.logger.info("Chinchilla instance created successfully.")

    @classmethod
    def from_config(cls, config_path: str, **kwargs) -> Chinchilla:
        """
        Constructs a Chinchilla instance from a configuration file, with the option to override specific settings.

        Args:
            config_path (str): The path to the configuration file in JSON or YAML format.
            **kwargs: Optional keyword arguments to override configuration settings.

        Returns:
            Chinchilla: A new instance of Chinchilla configured based on the provided file and overrides.

        Raises:
            ValueError: If the configuration file format is not supported.
        """
        with open(config_path, "r") as file:
            if config_path.lower().endswith(".json"):
                import json

                config = json.load(file)
            elif config_path.lower().endswith((".yaml", ".yml")):
                # import yaml  # Not compatible with the scientific notation like 1e18
                # config = yaml.safe_load(file)
                from ruamel.yaml import YAML

                yaml = YAML(typ="safe")

                config = yaml.load(file)
            else:
                raise ValueError("Unsupported file format. Please use JSON or YAML.")

        config.update(kwargs)
        return cls(**config)

    def _create_shortcuts(self) -> None:
        """Sets up shortcut methods."""
        # Submodules; consult each class for what it does
        self.append = self.database.append
        self.plot = lambda *args, **kwargs: self.visualizer.plot(self, *args, **kwargs)

        # Self-shorthands
        self.allocate = self.allocate_compute
        self.L = self.predict_loss
        self.__param_grid_keys = set(self.param_grid.keys())

    # Simulation utility
    def simulate(self, *args, **kwargs) -> None:
        """
        Simulates the scaling law estimation process using the provided arguments.
        This method is a wrapper around the Simulator class, allowing for quick setup and execution of simulations.

        Args:
            *args: Variable length argument list to be passed to `Simulator.__call__`.
            **kwargs: Arbitrary keyword arguments to be passed to `Simulator.__call__`.
        """
        from .simulator import Simulator

        Simulator(self)(*args, **kwargs)

    def seed(self) -> tuple[tuple[int, float], dict[str, int] | None]:
        """
        Sample a random allocation and model configuration from the user-specified seed regime.

        Returns:
            `(N, D), model_config` - A tuple containing the allocation $(N, D)$ followed by
            a model configuration dictionary corresponding to $N$. If `model_search_config` is not specified,
            the latter will be `None`.


        Raises:
            ValueError: If a valid configuration could not be found after a certain number of trials.
        """
        get_model_config = self.model_search_config is not None

        _max_iters = 2**10
        for i in range(_max_iters):
            C = np.exp(random.uniform(*np.log(self.seed_ranges.C)))
            ratio_N_to_D = np.exp(random.uniform(*np.log(self.seed_ranges.N_to_D)))
            N = np.sqrt(C / (6 * ratio_N_to_D))  # N^2 = C / (6*(D/N))
            D = N * ratio_N_to_D
            if get_model_config:
                model_config, N = search_model_config(N, **self.model_search_config, use_cache=i)
                C = 6 * N * D
            else:
                model_config = None
            if is_between(C, self.seed_ranges.C) and is_between(D / N, self.seed_ranges.N_to_D):
                break
        else:
            raise ValueError(f"We could not find a valid configuration in {_max_iters} trials.")

        self.logger.debug(f"[{ordinal(len(self.database.df)+1)}]\t{C:.2e} FLOPs => {N:.2e} params * {D:.2e} samples")

        return (N, D), model_config
        # return dict(N=N, D=D, model_config=model_config)

    def fit(self, parallel: bool = True, simulation: bool = False) -> None:
        """
        Uses [L-BFGS optimization (SciPy implementation)](https://docs.scipy.org/doc/scipy/reference/optimize.minimize-lbfgsb.html)
        to find the best-fitting parameters for the scaling law based on the collected data.

        Args:
            parallel (bool, optional): Whether to run L-BFGS optimization over the initialization grid in parallel processing.
            simulation (bool, optional): Indicates whether the fitting is part of a simulation. Defaults to False.

        Raises:
            ValueError: If there are not enough data points to perform the fitting.
            TypeError: If the numerical precision is insufficient for the L-BFGS algorithm.
            NotImplementedError: When you try to use `weight_fn` for the first time; you are supposed to start _hacking_ here.
        """
        _df = self.database.df.copy()

        if len(_df) < 3:
            raise ValueError(
                f"You do not have enough number of training runs on the database yet. Currently: {len(_df)}"
            )

        if DTYPE().itemsize < 8:  # In bytes
            raise TypeError(
                "The current operation requires a numerical precision of at least 64-bit as used in the L-BFGS algorithm. Lower precisions such as np.float32 or below are not supported for this operation. Please ensure you're using np.float64 or higher precision to avoid this error."
            )

        # Pre-compute the series repeatedly accessed by `self._evaluate_params`
        if self.weight_fn:
            raise NotImplementedError(
                "When specifying `weght_fn, you are expected to edit the source code by deleting this error and specify how to compute yourself."
            )
            weights = self.weight_fn(_df.C.values.astype(DTYPE))
            weights /= weights.mean()
        else:
            weights = None

        self._const = dict(
            log_N=np.log(_df.N.values.astype(DTYPE)),
            log_D=np.log(_df.D.values.astype(DTYPE)),
            y_true=_df.loss.values.astype(DTYPE),
            weights=weights,
        )

        # The absolute value range affects the differential optimization
        self._autoscale_range = np.array(list(map(np.ptp, self.param_grid.values())))
        # In case of any axis with a single initial value:
        self._autoscale_range[self._autoscale_range == 0] = 1.0

        if parallel:
            # GLobal declarations for `multiprocessing`
            global initial_guesses
            global _optimize_params

        initial_guesses = list(itertools.product(*self.param_grid.values()))
        initial_guesses /= self._autoscale_range

        def _optimize_params(i):
            x0 = initial_guesses[i]
            # res = sciop.minimize(self._evaluate_params, x0, method="L-BFGS-B", tol=1e-7)  # Note: `tol` -> `ftol`
            # lbfgs_loss = res.fun
            # if np.isfinite(lbfgs_loss):
            #     return res.x * self._autoscale_range, lbfgs_loss
            # https://github.com/scipy/scipy/blob/v1.12.0/scipy/optimize/_lbfgsb_py.py
            x, lbfgs_loss, _ = sciop.fmin_l_bfgs_b(
                self._evaluate_params,
                x0,
                approx_grad=True,
                maxiter=1_000_000,
                maxfun=1_000_000,
                # Default values generally perform fine
            )
            if np.isfinite(lbfgs_loss):
                return x, lbfgs_loss

        with multiprocessing.Pool(os.cpu_count()) as pool:
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                TimeElapsedColumn(),
                "/",
                TimeRemainingColumn(),
                disable=self.logger.getEffectiveLevel() > 30,
            ) as progress:
                task = progress.add_task(
                    ("" if not simulation else "[SIMULATION] ") + "Fitting scaling law", total=len(initial_guesses)
                )
                results = []
                if parallel:
                    self.logger.debug(f"{os.cpu_count()=}")
                    for res in pool.imap_unordered(_optimize_params, range(len(initial_guesses))):
                        if res:
                            results.append(res)
                        progress.update(task, advance=1.0)
                else:
                    for i in range(len(initial_guesses)):
                        res = _optimize_params(i)
                        if res:
                            results.append(res)
                        progress.update(task, advance=1.0)

        if not results:
            raise ValueError("No valid result from L-BFGS. `loss_fn` you have specified is possibly broken.")
        best_fit = min(results, key=lambda x: x[1])
        self.E, self.A, self.B, self.alpha, self.beta = best_fit[0] * self._autoscale_range

        # Adjust and update if any of the user-specified scales were logarithms
        for param_name in ("e", "a", "b"):
            if param_name in self.__param_grid_keys:
                setattr(self, param_name.upper(), np.exp(getattr(self, param_name.upper()), dtype=DTYPE))

        N, D = _df.N.values, _df.D.values
        # N, D = N.astype(float), D.astype(float)  # In case N and D were type object in pandas/numpy
        y_pred = self.predict_loss(N, D)
        self.visualizer.LBFGS(y_pred, _df.loss.values, simulation=simulation)

        self.logger.info(
            f"Loss predictor:\n\n  L(N, D) = {self.E:#.4g} + {self.A:#.4g} / (N ^ {self.alpha:#.4g}) + {self.B:#.4g} / (D ^ {self.beta:#.4g})\n"
        )

    def scale(
        self,
        scaling_factor: float | None = None,  # Optional override
        C: float | None = None,  # Directly specify the compute budget for the step
        simulation: bool = False,
    ) -> tuple[tuple[int, float], dict[str, int] | None]:
        """
        Determines the compute-optimal allocation of a scaled FLOP budget for the next model.

        > **Example Usages**:
        >
        > 1. Specifying/overriding `scaling_factor` in place
        > ```diff
        > cc = Chinchilla(
        >     ...
        > -    scaling_factor=10.0,
        >     ...
        > )
        > for i in range(100 + 8):
        >     ...
        >     cc.fit()
        > -    (N, D), model_config = cc.scale()
        > +    (N, D), model_config = cc.scale(scaling_factor=10.0)
        > ```
        >
        > 2. Directly specify the FLOP budget
        > ```python
        > (N, D), model_config = cc.scale(C=5.76e23)
        > ```

        Args:
            scaling_factor (float | None, optional): An optional scaling factor to override the instance's scaling factor. Defaults to None.
            C (float | None, optional): Directly specify the compute budget for the scaling step. Defaults to None.
            simulation (bool, optional): Indicates whether the scaling is part of a simulation. Defaults to False.

        Returns:
            `(N, D), model_config` - A tuple containing the allocation $(N, D)$ and
            an optional dictionary with the model configuration corresponding to $N$.

        Raises:
            ValueError: If any of the following conditions are met:
                - The scaling law parameters have not been estimated
                - neither `C` nor `scaling_factor` are specified
                - both `C` and `scaling_factor` are specified
        """
        get_model_config = self.model_search_config is not None

        if not all(hasattr(self, param) for param in ["alpha", "beta"]):
            raise ValueError("You must call `fit` before training a model with scaled compute.")

        if not any([C, scaling_factor, self.scaling_factor]):
            raise ValueError("You must specify either `C` or `scaling_factor`")

        if all([C, scaling_factor]):
            raise ValueError("You cannot specify both `C` and `scaling_factor`")

        # Prioritize parameters in order of C, scaling_factor, self.scaling_factor
        if C is None:
            # Use the preset `scaling_factor` if not overridden
            scaling_factor = scaling_factor or self.scaling_factor
            C = max(self.seed_ranges.C[1], int(self.database.df.C.max())) * scaling_factor

        N, D = self.allocate_compute(C)
        if get_model_config:
            model_config, N = search_model_config(N, **self.model_search_config)
            D = self.adjust_D_to_N(N)
            C = 6 * N * D
            if C <= int(self.database.df.C.max()):
                self.logger.warning("The scaling process has saturated. Consider increasing some numbers")
        else:
            model_config = None

        self.logger.info(f"[{ordinal(len(self.database.df)+1)}]\t{C:.2e} FLOPs => {N:.2e} params * {D:.2e} samples")
        self.plot(next_point=dict(C=C, N=N, D=D), simulation=simulation)

        return (N, D), model_config

    def step(
        self, num_seeding_steps: int | None = None, parallel: bool = True, simulation: bool = False, **scale_kwargs
    ) -> tuple[tuple[int, float], dict[str, int] | None]:
        """
        Shorthand method automatically routing to `seed` or `fit` & `scale` methods,
        depending on the existing number of training runs in the seed regime.

        > If you prefer to be explicit about the seeding and scaling steps, you can use the following approach:
        >
        > ```diff
        > - (N, D), model_config = cc.step(num_seeding_steps=401)
        > + if len(cc.database.df) < 401:
        > +     (N, D), model_config = cc.seed()
        > + else:
        > +     cc.fit_scaling_law()
        > +     (N, D), model_config = cc.scale()
        > ```

        Args:
            num_seeding_steps (int, optional): The threshold number of seed training runs before starting to scale the compute budget.
            parallel (bool, optional): Whether to run L-BFGS optimization over the initialization grid in parallel processing. To be passed to `fit`.
            simulation (bool, optional): Indicates whether the scaling is part of a simulation. Defaults to False.
            **scale_kwargs: Keyword arguments to be passed to `scale` (`scaling_factor` and `C`).

        Returns:
            `(N, D), model_config` - A tuple containing the allocation $(N, D)$ and
            an optional dictionary with the model configuration corresponding to $N.

        Raises:
            ValueError: If any of the following conditions are met:
                - `num_seeding_steps` is not specified
                - neither `C` nor `scaling_factor` are specified
                - both `C` and `scaling_factor` are specified
        """
        # Use the preset value if not overridden
        num_seeding_steps = num_seeding_steps or self.num_seeding_steps

        if num_seeding_steps is None:
            raise ValueError(
                "To use `step` shorthand method, you must specify `num_seeding_steps` either in `__init__` or this method"
            )

        num_valid_seeds = self.database.df.C.apply(lambda x: is_between(x, self.seed_ranges.C)).sum()
        if num_valid_seeds < num_seeding_steps:
            if not simulation:
                self.logger.info(
                    textwrap.dedent(
                        f"""\
                       Existing number of [i]seed[/] training runs: {num_valid_seeds}
                    -> Sample a model configuration (N, D) from the seed regime"""
                    )
                )
            return self.seed()
        else:
            # pre-validate scale_kwargs
            if not any([hasattr(scale_kwargs, "C"), hasattr(scale_kwargs, "scaling_factor"), self.scaling_factor]):
                raise ValueError("You must specify either `C` or `scaling_factor` as a positive number")

            if hasattr(scale_kwargs, "C") and hasattr(scale_kwargs, "scaling_factor"):
                raise ValueError("You cannot specify both `C` and `scaling_factor`")

            if not simulation:
                self.logger.info(
                    textwrap.dedent(
                        f"""\
                           Existing number of [i]seed[/] training runs: {num_valid_seeds}
                        -> Estimate the parameters ([i]A[/], [i]B[/], [i]α[/], [i]β[/]) from {len(self.database.df)} training runs"""
                    )
                )
            self.fit(simulation=simulation, parallel=parallel)
            return self.scale(simulation=simulation, **scale_kwargs)

    def adjust_D_to_N(self, N: float) -> float:
        r"""
        Adjusts $D$ (the number of data samples) to $N$ (the number of model parameters) based on the scaling law.
        Computes:

            $$D = G^{-(1 + b/a)} N^{b/a}$$

        > **Example Usage**:
        > ```python
        > (N, D), model_config = cc.scale()
        > model = Model(**model_config)
        > N = sum(p.numel() for p in model.parameters())
        > D = cc.adjust_D_to_N(N)
        > ```
        > Once you get an estimate of the scaling law for your task,
        > you may want to update $D$ to match the actual value of $N$ if your `estimate_model_size` is not strictly accurate.


        Args:
            N (float): The number of model parameters.

        Returns:
            float: The adjusted number of data samples.

        Raises:
            ValueError: If N is not a positive number.
        """
        if not isinstance(N, (int, float)) or N <= 0:
            raise ValueError(f"N must be a positive number, but got {N}")

        D = np.power(self.G, -1 - self._b / self._a) * np.power(N, self._b / self._a)

        self.logger.debug(f"Adjusted D to N: {D=:.1f} | {N=}")
        return D

    def allocate_compute(self, C: float | list | np.ndarray) -> tuple[float, float] | np.ndarray:
        r"""
        Allocates a given computational budget (C) to the optimal number of model parameters (N) and data samples (D):

        $$\underset{N,\ D}{argmin}\ L(N,\ D\ |\ E,\ A,\ B,\ \alpha,\ \beta)$$

        Args:
            C (float): The computational budget in FLOPs.

        Returns:
            tuple[float, float]: A tuple containing the optimal number of model parameters (N) and data samples (D).

        Raises:
            ValueError: If C is not a positive number.
        """
        alpha_beta = self.alpha + self.beta

        self.G = np.power((self.alpha * self.A) / (self.beta * self.B), (1 / alpha_beta))
        self._a, self._b = self.beta / alpha_beta, self.alpha / alpha_beta

        N_opt = self.G * np.power(C / 6, self._a)
        D_opt = (1 / self.G) * np.power(C / 6, self._b)

        return N_opt, D_opt

    def predict_loss(self, N: np.ndarray | float, D: np.ndarray | float) -> np.ndarray | float:
        """
        Predicts the loss for given allocations of model parameters (N) and data samples (D).

        Args:
            N (np.ndarray | float): The number of model parameters or an array of such numbers.
            D (np.ndarray | float): The number of data samples or an array of such numbers.

        Returns:
            *np.ndarray | float*: The predicted loss or an array of predicted losses.
        """
        log_term_2nd = np.log(self.A) - self.alpha * np.log(N)  # log (already done) -> sum exp
        log_term_3rd = np.log(self.B) - self.beta * np.log(D)
        return self.E + np.exp(log_term_2nd) + np.exp(log_term_3rd)

    def _evaluate_params(self, x) -> np.ndarray:
        """
        Internal method to compute the loss for the L-BFGS algorithm.

        This method evaluates the loss function for a given set of parameters during the optimization process.

        Args:
            X (np.ndarray): The array of parameters to evaluate.

        Returns:
            float: The computed loss value.

        Raises:
            ValueError: If NaN values are detected in the input parameters.
        """
        if np.isnan(x).any():
            raise ValueError(
                f"NaN value(s) detected in input to `cc._evaluate_params`: {x}\n"
                f"This was possibly because `loss_fn` you specified is compatible with type `{DTYPE}`"
            )

        # Scipy/Fortran implementation of LBFGS casts `x0` to float64 internally, so recover here.
        # Invert autoscaling & decompose
        E, a, b, alpha, beta = x.astype(DTYPE) * self._autoscale_range

        # Ensure the log scale for `a` and `b` but `E`.
        # Inspect the user-specified initial keys requiring `exp`/`log` transformation
        if "e" in self.__param_grid_keys:
            E = np.exp(E)
        if "A" in self.__param_grid_keys:
            a = np.log(np.clip(a, FLOAT_TINY, None))
        if "B" in self.__param_grid_keys:
            b = np.log(np.clip(b, FLOAT_TINY, None))

        log_term_2nd = a - alpha * self._const["log_N"]
        log_term_3rd = b - beta * self._const["log_D"]
        log_term_2nd = np.clip(log_term_2nd, None, FLOAT_LOGMAX)
        log_term_3rd = np.clip(log_term_3rd, None, FLOAT_LOGMAX)
        y_pred = E + np.exp(log_term_2nd) + np.exp(log_term_3rd)

        losses = self.loss_fn(self._const["y_true"], y_pred)

        if self.weight_fn:
            losses = losses * self._const["weights"]

        return np.mean(losses)

    def get_params(self) -> dict:
        """
        Returns a dictionary of estimated parameters describing the scaling law / parametric loss estimator.

        Returns:
            float: The computed loss value.

        Raises:
            ValueError: If the scaling law parameters have not been estimated
        """
        if not all(hasattr(self, param) for param in ["alpha", "beta"]):
            raise ValueError("You must call `fit` before training a model with scaled compute.")
        return {
            "E": self.E,
            "A": self.A,
            "B": self.B,
            "alpha": self.alpha,
            "beta": self.beta
        }

    def report(self, plot: bool=True) -> None:
        """
        Generates a report summarizing the scaling law estimation results.

        The report includes:
        - Estimated scaling law parameters (E, A, B, alpha, beta)
        - Goodness of fit with regards to specified loss function
        - Plots of the actual vs. predicted losses
        - Suggested compute allocation for the next model based on the scaling law

        Raises:
            ValueError: If the scaling law parameters have not been estimated yet.
        """
        if not all(hasattr(self, param) for param in ["alpha", "beta"]):
            raise ValueError("You must call `fit` before generating a report.")

        self.logger.info("Estimated scaling law parameters:")
        for param, value in self.get_params().items():
            self.logger.info(f"    - {param}: {value}")
        if len(self.database.df):
            self.logger.info("Goodness of fit:")
            y_true = self.database.df.loss.values
            y_pred = self.predict_loss(self.database.df.N.values, self.database.df.D.values)
            weights = 1.0
            if self.weight_fn:
                weights = self.weight_fn(self.database.df.C.values.astype(DTYPE))
                weights /= weights.mean()
            loss = (self.loss_fn(y_true, y_pred) * weights).mean()
            self.logger.info(f"    - `{self.loss_fn.__name__}`: {loss:.8f}")
            if plot:
                self.logger.info("Landscape visualization:")
                self.plot()



"""Utility functions."""
from __future__ import annotations  # PEP 604 backport

import itertools
from collections.abc import Callable
import numpy as np

# Global variable to store cached estimates for model configurations.
# This should be used with caution and only when the hyperparameter grid is consistent:
# e.g., inside a loop to find a valid configuration.
cached_estimates: list[tuple[dict[str, float], float]] | None = None


def search_model_config(
    N_opt: float,
    hyperparam_grid: dict[str, list[float]],
    size_estimator: Callable[..., float],
    use_cache: bool = False,
) -> tuple[dict[str, float], float]:
    """
    Finds the model configuration that is closest to a given target number of parameters,
    based on the provided hyperparameter grid and size estimator.

    > **Example Usage**:
    > ```python
    > from chinchilla._utils import search_model_config
    > def my_size_estimator_function(hidden_size, num_layers):
    >     # ... Write your code to estimate `N`
    >
    > model_search_config = {
    >     hyperparam_grid={
    >         'hidden_size': [1024, 2048, 4096],
    >         'num_layers': [1, 2, 4, 8, 16, 32]
    >     },
    >     size_estimator=my_size_estimator_function
    > }
    > model_config, estimated_size = search_model_config(1e9, **model_search_config)
    > ```

    Args:
        N_opt (float): The target number of parameters for the model configuration.
        hyperparam_grid (dict[str, list[float]]): A dictionary where keys are hyperparameter names and values are lists of possible values.
        size_estimator (Callable[..., float]): A callable that takes a model configuration as keyword arguments and returns the estimated size.
        use_cache (bool, optional): A boolean flag indicating whether to use cached estimates. Default is False.

    Returns:
        A tuple containing the closest model configuration and its estimated size.

    Note:
        Although very efficient, you should set `use_cache` to True only when `hyperparam_grid` is guaranteed to be
        consistent; thus, it is disabled by default except for Simulator (x16 faster).
    """
    global cached_estimates
    if not use_cache or cached_estimates is None:
        # Pre-compute and sort estimates if not already done
        estimates = []
        for _values in itertools.product(*hyperparam_grid.values()):
            config = dict(zip(hyperparam_grid.keys(), _values))
            N = size_estimator(**config)
            if N and np.isfinite(N):
                estimates.append((config, N))
        estimates.sort(key=lambda x: x[1])  # Sort by estimated size
        cached_estimates = estimates

    closest = cached_estimates[np.abs(np.array([e[1] for e in cached_estimates]) - N_opt).argmin()]
    return closest


def is_between(value: float | np.ndarray, bounds: tuple[float, float] | np.ndarray) -> bool | np.ndarray:
    """
    Checks if a value is within the given inclusive bounds.

    Args:
        value (float | np.ndarray): The value to check.
        bounds (tuple[float, float] | np.ndarray): A tuple containing the lower and upper bounds.

    Returns:
        bool | np.ndarray: NumPy array: A boolean or an NumPy array of booleans indicating whether the value is between the bounds.
    """
    lower, upper = bounds
    return lower <= value <= upper



"""
Contains a utility function `get_logger`. This module also filters out noisy debug messages 
from `matplotlib` and suppresses redundant warnings from `numpy` and `matplotlib`.
"""

from __future__ import annotations  # PEP 604 backport

import logging
import warnings

from rich.console import Console
from rich.logging import RichHandler

# Ignore the noisy matplotlib debug messages.
logging.getLogger("matplotlib").setLevel(logging.ERROR)

# For NumPy's precision errors & Matplotlib with non-GUI backend
warnings.filterwarnings("once", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=UserWarning)


def get_logger(level: int | str, name: str) -> logging.Logger:
    """
    Sets up a logger with the specified log level.
    This logger uses RichHandler for `rich` formatted logging output to the console.

    Args:
        level (int | str): Logging level, e.g., 20 or logging.INFO, 30 or logging.WARNING.
        name (str, optional): The name of the logger.

    Returns:
        logging.Logger: Configured logger instance.
    """
    # Configure the logger for this library
    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(console=Console(stderr=True), markup=True)],
        force=True,
    )
    # Hook warnings to logger
    logging.captureWarnings(True)
    return logging.getLogger(name)



"""A few loss & weight functions you can use on demand."""
from __future__ import annotations  # PEP 604 backport

import numpy as np


def asymmetric_mae(y_true: np.ndarray, y_pred: np.ndarray, w: float = 1e1) -> np.ndarray:
    r"""Asymmetric Mean Absolute Error loss function."""
    error = y_true - y_pred
    error[error < 0] *= -w  # abs() and weight as a time
    return error


def huber(y_true: np.ndarray, y_pred: np.ndarray, delta: float = 1.0) -> np.ndarray:
    """Huber loss function."""
    error = y_true - y_pred
    abs_err = np.abs(error)
    loss = np.where(abs_err < delta, np.power(error, 2) / 2, delta * (abs_err - delta / 2))
    return loss


def log_huber(y_true: np.ndarray, y_pred: np.ndarray, delta: float = 1.0) -> np.ndarray:
    """The original loss function used in the Chinchilla paper"""
    y_true, y_pred = np.log(y_true), np.log(y_pred)
    return huber(y_true, y_pred, delta)


def mae(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """Mean Absolute Error loss function."""
    return np.abs(y_true - y_pred)


def mse(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """Mean Squared Error loss function."""
    return np.power(y_true - y_pred, 2)



from __future__ import annotations  # PEP 604 backport

import random
from typing import Callable, Iterator, Any

import numpy as np

from rich.progress import track

from ._logger import get_logger
from ._validator import SimulationArgs
from .core import Chinchilla
from .database import Database


class Simulator(Chinchilla):
    """
    Simulates the scaling law estimation with `Chinchilla`, allowing you to understand its behaviors.

    Inheriting and extending the `Chinchilla` class with the capacity to simulate seeding and scaling in a hypothetical task,
    `Simulator` models how factors like `Chinchilla` configuration, number of seeds, scaling factor, the noisiness of losses, etc.
    would confound to affect the stability and the performance of scaling law estimation.

    Attributes:
        cc (Chinchilla): The Chinchilla instance with preset configuration attributes like `param_grid`, `seed_ranges`, `loss_fn`, etc.
        logger (Logger): Logger instance for simulation activities.
        database (Database): Database instance to record simulation results.

    Methods:
        __init__: Initializes the Simulator with a Chinchilla instance.
        __getattr__: Delegates attribute access to the Chinchilla instance.
        __call__: Executes the simulation with given parameters.
        _pseudo_training_run: Performs a pseudo-training run and records results.
    """

    def __init__(self, cc: Chinchilla):
        """
        Initialize the Simulator with a Chinchilla instance.

        Args:
            cc (Chinchilla): An instance of Chinchilla to be used for simulation.
        """
        self.cc = cc
        self.log_level = self.cc.logger.getEffectiveLevel()
        self.logger = get_logger(self.log_level, name="chinchilla.simulator")
        self._create_shortcuts()

    def __getattr__(self, name: str) -> Any:
        """
        Delegate attribute access to the Chinchilla instance.

        Args:
            name: The name of the attribute to access.

        Returns:
            Any: The value of the attribute with the specified name from the Chinchilla instance.
        """
        return getattr(self.cc, name)

    def __call__(
        self,
        num_seeding_steps: int = 100,
        num_scaling_steps: int = 1,
        scaling_factor: float | None = None,
        target_params: dict[str, float] = {
            "E": 1.69337368,
            "A": 406.401018,
            "B": 410.722827,
            "alpha": 0.33917084,
            "beta": 0.2849083,
        },
        noise_generator: Iterator | tuple[Callable, tuple[float, ...]] | None = (
            random.expovariate(10) for _ in iter(int, 1)
        ),
    ) -> None:
        """
        Simulate the compute-scaling on a hypothetical deep learning task with some noise expectable from reality.

        Args:
            num_seeding_steps (int): The number of seeding steps to simulate.
            num_scaling_steps (int): The number of scaling steps to simulate.
            scaling_factor (float | None, optional): The scaling factor to be used in the simulation.
            target_params (dict[str, float]): A dictionary of target parameters for the simulation.
            noise_generator (Iterator | tuple[Callable, tuple[float, ...]] | None, optional): A callable or iterator that generates noise to be added to the loss.
                Defaults to `(random.expovariate(10) for _ in iter(int, 1))`, which generates an exponential distribution averaging at $0.100$.

        Raises:
            TypeError: If the provided `noise_generator` is not an iterator or a tuple with a callable and its arguments.
        """
        # Preprocessing: if `noise_generator` is passed as a pair of method and args, convert into an infinite generator
        # e.g., `(random.expovariate, (10,))`
        if isinstance(noise_generator, (tuple, list)):
            func, args = noise_generator
            if not (callable(func) and isinstance(args, (tuple, list))):
                raise TypeError(
                    "When passing a callable and arguments as `noise_generator`, it must be executable as `func(*args)`"
                )
            noise_generator = (func(*args) for _ in iter(int, 1))

        # Validation
        SimulationArgs(
            num_seeding_steps=num_seeding_steps,
            num_scaling_steps=num_scaling_steps,
            scaling_factor=scaling_factor or self.scaling_factor,
            target_params=target_params,
            noise_generator=noise_generator,
        )
        self.num_seeding_steps = num_seeding_steps
        self.num_scaling_steps = num_scaling_steps
        self.scaling_factor = scaling_factor
        self.target_params = target_params
        self.noise_generator = noise_generator

        # Reset with an in-memory database *not* saving appended data.
        self.database = Database(log_level=self.log_level)
        self.logger.info("[b]Starting a simulation[/]")
        # Main: simulate seeding and scaling
        for _ in track(
            range(self.num_seeding_steps),
            description="  [SIMULATION] Seed training runs ",
            disable=self.log_level <= 10 or 30 < self.log_level,  # logging.WARNING
        ):
            self._pseudo_training_run()

        self.logger.debug("Data types:\n" + self.database.df.dtypes.T.to_string())

        for _ in range(self.num_scaling_steps):  # Logger will show the progress
            self._pseudo_training_run()

        if not num_scaling_steps:
            # Plot the distribution of seed models with the estimated loss gradient
            self.plot(simulation=True)

    def _pseudo_training_run(self) -> None:
        """Perform a pseudo training run and record the results in the database."""
        (N, D), _ = self.step(simulation=True)

        # Get a *hypothetical* lowest loss you can achieve with N and D
        pseudo_loss = sum(
            [
                self.target_params["E"],
                np.exp(np.log(self.target_params["A"]) - self.target_params["alpha"] * np.log(N)),
                np.exp(np.log(self.target_params["B"]) - self.target_params["beta"] * np.log(D)),
            ]
        )
        # Add a random noise (>0.0) attributed to inefficiency of training setup
        if self.noise_generator:
            pseudo_loss += next(self.noise_generator)

        self.database.append(N=N, D=D, C=6 * N * D, loss=pseudo_loss)  # type: ignore
